local m, s, o
local rule_opt = {}
local redir_opt = {}
local vpn_type = "IPsec"

local uci = require("luci.model.uci").cursor()
local util_tlt = require "luci.util_tlt"

rule_opt[1] = {
	name 		= "Allow-"..vpn_type.."-ESP",
	target 		= "ACCEPT",
	src  		= "wan",
	proto 		= "esp",
	vpn_type	= vpn_type
}

rule_opt[2] = {
	name 		= "Allow-"..vpn_type.."-NAT-T",
	target 		= "ACCEPT",
	src  		= "wan",
	proto 		= "udp",
	dest_port	= "4500",
	vpn_type	= vpn_type
}

rule_opt[3] = {
	name 		= "Allow-"..vpn_type.."-IKE",
	target 		= "ACCEPT",
	src  		= "wan",
	dest_port	= "500",
	proto 		= "udp",
	vpn_type	= vpn_type
}

rule_opt[4] = {
	name 		= "Allow-"..vpn_type.."-Forward",
	target 		= "ACCEPT",
	src  		= "wan",
	proto 		= "all",
	dest 		= "*",
	extra 		= '-m policy --dir in --pol ipsec',
	vpn_type	= vpn_type
}

redir_opt[1] = {
	name		= "Exclude-"..vpn_type.."-from-NAT",
	proto		= 'any',
	extra		= '-m policy --dir out --pol ipsec',
	dest		= 'wan',
	target 		= "ACCEPT",
	vpn_type	= vpn_type
}

m = Map("ipsec")
m.redirect = luci.dispatcher.build_url("admin", "services", "vpn", "ipsec")
if not m.uci:get(m.config, arg[1]) then
	luci.http.redirect(m.redirect)
end
--Remote section

s = m:section(NamedSection, arg[1], "remote", translatef("IPsec instance: %s", arg[1]), translate("This section is used to configure the settings of the IPsec instance. Scroll your mouse pointer over field names in order to see helpful hints."))

s:tab("general", translate("General Settings"))
s:tab("advanced", translate("Advanced Settings"))

o = s:taboption("general", Flag, "enabled", translate("Enable"), translate("Turns the IPsec instance on or off."))
o.rmempty = false

o = s:taboption("general", Value, "gateway", translate("Remote endpoint"), translate("Domain name or IP address. Leave empty to only accept connections"))
o.datatype = "host"

o = s:taboption("general", ListValue, "authentication_method", translate("Authentication method"), translate("Choose the authentication. Pre-sahred key or x.509 standart certificates"))
o:value("psk", "Pre-shared key")
o:value("x509", "X.509")
o.rmempty = false

o = s:taboption("general", Value, "pre_shared_key", translate("Pre shared key"), translate("A shared password used for authentication between IPsec peers. Allowed characters (a-zA-Z0-9!@#$%&*+-/=?^_`{|}~. )"))
o.password = true
o.datatype = "fieldvalidation('^[a-zA-Z0-9!@#$%%&*+/=?^_`{|}~.-]+$',0)"
o:depends("authentication_method", "psk")

o = s:taboption("general", FileUpload, "key", "Key", "Upload private key file")
o:depends("authentication_method", "x509")
o.unsafeupload = true

o = s:taboption("general", FileUpload, "leftcert", "Local certificate", "Upload this device certificate file")
o:depends("authentication_method", "x509")
o.unsafeupload = true

o = s:taboption("advanced", FileUpload, "rightcert", "Remote certificate", "Upload remote device certificate file")
o:depends("authentication_method", "x509")
o.unsafeupload = true

o = s:taboption("general", FileUpload, "cacert", "CA Certificate", "Upload CA certificate file")
o:depends("authentication_method", "x509")
o.unsafeupload = true

o = s:taboption("general", Value, "local_identifier", translate("Local identifier"),translate("How the left participant should be identified for authentication"))
o.placeholder = "IP, FQDN"
o.datatype = "string"

o = s:taboption("general", Value, "remote_identifier", translate("Remote identifier"), translate("How the right participant should be identified for authentication"))
o.placeholder = "IP, FQDN"
o.datatype = "string"

--Connection section

s = m:section( NamedSection, arg[1].."_c", "connection", translate("Connection settings"))

s:tab("general", translate("General Settings"))
s:tab("advanced", translate("Advanced Settings"))

o = s:taboption("general", ListValue, "mode", translate("Mode"), translate("What operation should be done automatically at IPsec startup"))
o:value("start", "Start")
o:value("add", "Add")
o:value("route", "Route")

o = s:taboption("general", ListValue, "type", translate("Type"), translate("Type of connection.".."<br />"..
"Tunnel - protects internal routing information by encapsulating the entire IP packet (IP header and payload); is commonly used in site-to-site VPN connections; supports NAT traversal.".."<br />"..
"Transport - only encapsulates IP payload data; is used in client-to-site VPN connections; does not support NAT traversal; usually implemented with other tunneling protocols (for example, L2TP)."))
o:value("tunnel", "Tunnel")
o:value("transport", "Transport")

function o.write(self, section, value)
	--~ Add list to remote section for config generation
	m.uci:delete(m.config, arg[1], "transport")
	m.uci:delete(m.config, arg[1], "tunnel")
	m.uci:set_list(m.config, arg[1], value, section)
	return self.map:set(section, self.option, value)
end

o = s:taboption("general", DynamicList, "local_subnet", translate("Local subnet"), translate("Private subnet behind the left (local) participant, expressed as network/netmask. IKEv2 supports multiple subnets, IKEv1 only interprets the first subnet"))
o.datatype = "list(netmask)"
o.placeholder = "192.168.1.0/24"
o:depends({type="tunnel",defaultroute=0})

o = s:taboption("general", DynamicList, "remote_subnet", translate("Remote subnet"), translate("Private subnet behind the right (remote) participant, expressed as network/netmask. IKEv2 supports multiple subnets, IKEv1 only interprets the first subnet"))
o.datatype = "list(netmask)"
o.placeholder = "192.168.2.0/24"
o:depends({type="tunnel", defaultroute=0})

o = s:taboption("general", ListValue, "bind_to", translate("Bind to"), translate("Bind to GRE interface to create GRE over IPsec"))
o:depends("type", "transport")
o:value("-", "None")

m.uci:foreach("network", "interface",
	function(s)
		if s.proto == "gre" then
			o:value(s['.name'], s['.name'].." (GRE)")
		end
	end
)

--~ Get interface option value before write to be able to delete gre auto
local interface_name = o:formvalue(s.section)
function o.write(self, section, value)
	if value ~= "-" then
		uci:set("network", value, "auto", "0")
	else
		local interface_name = self.map:get(section, self.option)
		uci:delete("network", interface_name, "auto")
	end
	return self.map:set(section, self.alias or self.option, value)
end

o = s:taboption("general", ListValue, "keyexchange", translate("Key exchange"), translate("Internet Key Exchange (IKE) version used for exchanging keys.".."</br>"..
"IKEv1 - more commonly used but contains known issues, for example, dealing with NAT.".."</br>"..
"IKEv2 - updated version with increased and improved capabilities, such as integrated NAT support, supported multihosting, deprecated exchange modes (does not use main or aggressive mode; only 4 messages required to establish a connection)."))
o:value("ikev1", "IKEv1")
o:value("ikev2", "IKEv2")

o = s:taboption("general", Flag, "defaultroute", translate("Default route"), translate("Route all traffic through the IPSec tunnel"))
o.rmempty = false
o:depends({type="tunnel"})

o = s:taboption("advanced", Flag, "aggressive", translate("Aggressive"), translate("Aggressive can be used only on outgoing connections"))
o:depends("keyexchange", "ikev1")

function o.write(self, section, value)
	local new_value
	if value == "1" then
		new_value = "yes"
	else
		new_value = ""
	end
	return AbstractValue.write(self, section, new_value)
end

function o.cfgvalue(self, section)
	local value = m:get(section, self.option)
	if value == "yes" then return "1" end
end

o = s:taboption("advanced", Flag, "forceencaps", translate("Force encapsulation"), translate("Force UDP encapsulation for ESP packets even if no NAT situation is detected. This may help to surmount restrictive firewalls"))
o.enabled = "yes"
o.disabled = "no"

o = s:taboption("advanced", Flag, "local_firewall", translate("Local firewall"), translate("Whether the left (local) participant is doing forwarding-firewalling (including masquerading) using iptables for traffic from the right (remote) subnet"))
o.enabled = "yes"
o.disabled = "no"
o.default = "yes"
o.rmempty = false

o= s:taboption("advanced", Flag, "remote_firewall", translate("Remote firewall"), translate("Whether the right (remote) participant is doing forwarding-firewalling (including masquerading) using iptables for traffic from the left (local) subnet"))
o.enabled = "yes"
o.disabled = "no"
o.rmempty = false

o = s:taboption("advanced", Value, "inactivity", translate("Inactivity"), translate("Defines the timeout interval, after which a CHILD_SA is closed if it did not send or receive any traffic"))
o.placeholder = "none"
o.datatype = "uinteger"

o = s:taboption("advanced", Flag, "_dpd", translate("Dead peer detection"), translate("Controls the use of the Dead Peer Detection protocol where notification messages are periodically sent in order to check the liveliness of the IPsec peer"))

o = s:taboption("advanced", ListValue, "dpdaction", translate("DPD action"), translate("controls the use of the Dead Peer Detection protocol where notification messages\
 are periodically sent in order to check the liveliness of the IPsec peer"))
o:depends("_dpd",1)
o:value("restart", "Restart")
o:value("hold", "Hold")
o:value("clear", "Clear")
o:value("none", "None")

o = s:taboption("advanced", Value, "dpddelay", translate("DPD delay"), translate("defines the period time interval with which R_U_THERE messages/INFORMATIONAL exchanges are sent to the peer"))
o.placeholder = "30s"
o:depends("_dpd",1)

--allows "%config" value
o = s:taboption("advanced", Value, "remote_sourceip", translate("Remote source IP"), translate("The internal source IP to use in a tunnel for the remote (right) peer"))
o.datatype = "or(ipaddr,fieldvalidation('^[%%config]+$',0))"
o.validator_hint = "IPv4 and IPv6 addresses are accepted. E.g. 192.168.1.1 . Phrase \"%config\" is also accepted"
o.placeholder = "10.0.2.0/24"
o:depends("type", "tunnel")

--allows "%config" value
o = s:taboption("advanced", Value, "local_sourceip", translate("Local source IP"), translate("The internal (left) source IP to use in a tunnel, also known as virtual IP"))
o.datatype = "or(ipaddr,fieldvalidation('^[%%config]+$',0))"
o.validator_hint = "IPv4 and IPv6 addresses are accepted. E.g. 192.168.1.1 . Phrase \"%config\" is also accepted"
o.placeholder = "10.0.1.0/24"
o:depends("type", "tunnel")

o = s:taboption("advanced", DynamicList, "rightdns", translate("Remote DNS"), translate("Comma separated list of DNS server addresses to exchange as configuration attributes. On the responder, only fixed IPv4/IPv6 addresses are allowed and define DNS servers assigned to the client"))
o.placeholder = "8.8.8.8"
o.datatype = "ipaddr"
o:depends("type", "tunnel")

o = s:taboption("advanced", Value, "leftprotoport", translate("Locally allowed protocols"), translate("Allowed protocols and ports over connection, also called Port Selectors. Defines in form of 'protocol/port' eg:'17/1701' or '17/%any' or 'udp/l2f'"))
o.placeholder = "tcp/smtp"
o.datatype = "fieldvalidation('^[a-zA-Z0-9/%%]+$',0)"

o = s:taboption("advanced", Value, "rightprotoport", translate("Remotely allowed protocols"), translate("Allowed protocols and ports over connection, also called Port Selectors"))
o.placeholder = "tcp/smtp"
o.datatype = "fieldvalidation('^[a-zA-Z0-9/%%]+$',0)"

o = s:taboption("advanced", DynamicList, "custom", translate("Custom option"), translate("Syntax opton=value"))
o.placeholder = "reqid=1"
o.datatype = "fieldvalidation('^[a-zA-Z0-9=]+$',0)"

net = s:taboption("advanced", Value, "passthrough", translate("Passthrough"), translate("Select passthrough interfaces"))
net.template = "cbi/network_netlist"
net.widget = "checkbox"
net.nocreate = true
net.cast = "string"
net:depends({type="tunnel", ["ipsec."..arg[1].."_c"..".defaultroute"] = 0})

function net.write(self, section, value)
	local values = ""
	if type(value) == "table" then
		for _, iface in pairs(value) do
			values = values .. iface .. " "
		end
		values = values:sub(1, #values - 1)
	else
		values = value
	end
	uci:set(self.config, section, "passthrough", values)
end


--Proposal section

s = m:section( NamedSection, arg[1].."_ph1", "proposal", translate("Proposal settings"), translate("This section is used to configure IKE (Internet Key Exchange) phase 1 & 2 settings. IKE is a protocol used to set up security associations (SAs) for the IPsec connection. This process is required before any IPsec tunnel can be established."))

s:tab("phase1",  translate("Phase 1"))
s:tab("phase2", translate("Phase 2"))

o = s:taboption("phase1", ListValue, "encryption_algorithm", translate("Encryption algorithm"),
	translate("Algorithm used for data encryption."))
o.default = "aes128"
o:value("3des","3DES")
o:value("aes128","AES 128")
o:value("aes192","AES 192")
o:value("aes256","AES 256")
o:value("aes128gcm8","AES128 GCM8")
o:value("aes192gcm8","AES192 GCM8")
o:value("aes256gcm8","AES256 GCM8")
o:value("aes128gcm12","AES128 GCM12")
o:value("aes192gcm12","AES192 GCM12")
o:value("aes256gcm12","AES256 GCM12")
o:value("aes128gcm16","AES128 GCM16")
o:value("aes192gcm16","AES192 GCM16")
o:value("aes256gcm16","AES256 GCM16")

o = s:taboption("phase1", ListValue, "hash_algorithm", translate("Authentication"),
	translate("Algorithm used for exchanging authentication information."))
o.default = "sha1"
o:value("md5", "MD5")
o:value("sha1", "SHA1")
o:value("sha256", "SHA256")
o:value("sha384", "SHA384")
o:value("sha512", "SHA512")

o = s:taboption("phase1", ListValue, "dh_group", translate("DH group"),
	translate("The DH (Diffie-Hellman) group must match with another incoming connection to establish IPSec"))
o.default = "modp1536"
o:value("modp768", "MODP768")
o:value("modp1024", "MODP1024")
o:value("modp1536", "MODP1536")
o:value("modp2048", "MODP2048")
o:value("modp3072", "MODP3072")
o:value("modp4096", "MODP4096")
o:value("ecp192", "ECP192")
o:value("ecp224", "ECP224")
o:value("ecp256", "ECP256")
o:value("ecp384", "ECP384")
o:value("ecp521", "ECP521")

o = s:taboption("phase1", Flag, "force_crypto_proposal", translate("Force crypto proposal"), translate("Only chosen proposals will be used"))
o.rmempty = false
function o.cfgvalue(self, section)
	return self.map:get(arg[1], self.option)
end

function o.write(self, section, value)
	m:set(arg[1], self.option, value)
end

o = s:taboption("phase1", Value, "ikelifetime", translate("IKE lifetime"), translate("How long the keying channel of a connection (ISAKMP or IKE SA) should last before being renegotiated. The time is specified in seconds. The s, m, h and d suffixes explicitly define the units for seconds, minutes, hours and days, respectively"))
o.placeholder = "3h"
o.datatype = "lengthvalidation(0,7,'^[0-9]+[smhd]?$',0)"

function o.cfgvalue(self, section)
	return self.map:get(arg[1].."_c", self.option)
end

function o.write(self, section, value)
	m:set(arg[1].."_c", self.option, value)
end

--~ Phase 2 settings
o = s:taboption("phase2", ListValue, "encryption_algorithm2", translate("Encryption algorithm"),
	translate("Algorithm used for data encryption."))
o.default = "aes128"
o:value("3des","3DES")
o:value("aes128","AES 128")
o:value("aes192","AES 192")
o:value("aes256","AES 256")
o:value("aes128gcm8","AES128 GCM8")
o:value("aes192gcm8","AES192 GCM8")
o:value("aes256gcm8","AES256 GCM8")
o:value("aes128gcm12","AES128 GCM12")
o:value("aes192gcm12","AES192 GCM12")
o:value("aes256gcm12","AES256 GCM12")
o:value("aes128gcm16","AES128 GCM16")
o:value("aes192gcm16","AES192 GCM16")
o:value("aes256gcm16","AES256 GCM16")

function o.cfgvalue(self, section)
	return self.map:get(arg[1].."_ph2", self.option:sub(1,-2))
end

function o.write(self, section, value)
	m:set(arg[1].."_ph2", self.option:sub(1,-2), value)
end

o = s:taboption("phase2", ListValue, "hash_algorithm2", translate("Hash algorithm"),
	translate("Algorithm used for exchanging hash information."))
o.default = "sha1"
o:value("md5", "MD5")
o:value("sha1", "SHA1")
o:value("sha256", "SHA256")
o:value("sha384", "SHA384")
o:value("sha512", "SHA512")

function o.cfgvalue(self, section)
	return self.map:get(arg[1].."_ph2", self.option:sub(1,-2))
end

function o.write(self, section, value)
	m:set(arg[1].."_ph2", self.option:sub(1,-2), value)
end

o = s:taboption("phase2", ListValue, "dh_group2", translate("PFS group"),
	translate("The PFS (Perfect Forward Secrecy) group must match with another incoming connection to establish IPSec"))
o.default = "modp1536"
o:value("modp768", "MODP768")
o:value("modp1024", "MODP1024")
o:value("modp1536", "MODP1536")
o:value("modp2048", "MODP2048")
o:value("modp3072", "MODP3072")
o:value("modp4096", "MODP4096")
o:value("ecp192", "ECP192")
o:value("ecp224", "ECP224")
o:value("ecp256", "ECP256")
o:value("ecp384", "ECP384")
o:value("ecp521", "ECP521")
o:value("no_pfs", "No PFS")

function o.cfgvalue(self, section)
	return self.map:get(arg[1].."_ph2", self.option:sub(1,-2))
end

function o.write(self, section, value)
	m:set(arg[1].."_ph2", self.option:sub(1,-2), value)
end


o = s:taboption("phase2", Flag, "force_crypto_proposal2", translate("Force crypto proposal"), translate("Only chosen proposals will be used"))
o.rmempty = false
function o.cfgvalue(self, section)
	return self.map:get(arg[1].."_c", self.option:sub(1,-2))
end

function o.write(self, section, value)
	m:set(arg[1].."_c", self.option:sub(1,-2), value)
end

o = s:taboption("phase2", Value, "lifetime", translate("Lifetime"), translate("How long a particular instance of a connection (a set of encryption/authentication keys for user packets) should last, from successful negotiation to expiry. The time is specified in seconds. The s, m, h and d suffixes explicitly define the units for seconds, minutes, hours and days, respectively"))
o.placeholder = "1h"
o.datatype = "lengthvalidation(0,7,'^[0-9]+[smhd]?$',0)"

function o.cfgvalue(self, section)
	return self.map:get(arg[1].."_c", self.option)
end

function o.write(self, section, value)
	m:set(arg[1].."_c", self.option, value)
end

m.on_commit = function (self)
	local enabled  = false
	local exists  = false
	local enabled_redir  = false
	local sid

	uci:foreach("ipsec", "remote", function(s)
		if s.enabled == "1" then
			enabled = true
		end
	end)

	uci:foreach("firewall", "redirect", function(s)
		if s.vpn_type == vpn_type then
			exists  = true
			sid = s[".name"]
			if s.enabled_redir == "1" then
				enabled_redir = true
			end
		end
	end)

	if enabled then
		for i=1,#rule_opt do
			luci.util_tlt.enable_vpn_firewall(vpn_type, nil, nil, rule_opt[i])
		end

		if not exists then
			for i=1,#redir_opt do
				uci:section("firewall", "redirect", nil, redir_opt[i])
			end
		elseif not enabled_redir then
			uci:set("firewall", sid, "enabled", "1")
		end
	else
		luci.util_tlt.disable_vpn_firewall(vpn_type)
	end
end

return m
